from django.contrib.gis.db.models import PointField
from django.db import models


class Restaurant(models.Model):
    name = models.CharField(max_length=200)
    location = PointField(null=True)

    def __str__(self):
        return self.name



from pathlib import Path

import pytest
from django.contrib.gis.geos import Point

from tests.test_gisserver.models import Restaurant

HERE = Path(__file__).parent
SRID_RD_NEW = 28992


@pytest.fixture()
def restaurant() -> Restaurant:
    return Restaurant.objects.create(
        name="Café Noir", location=Point(122411, 486250, srid=SRID_RD_NEW)
    )



"""Exceptions for the WFS operations.

Not all exception codes are listed here, only the one that
apply to the current set of supported operations.
"""
from django.template.loader import render_to_string


class OWSException(Exception):
    """Base class for XML based exceptions in this module."""

    status_code = 400  # Most common code in spec
    service = None
    version = "2.0.0"
    code = None
    text_template = None

    def __init__(self, locator, text=None, code=None):
        text = text or self.text_template.format(code=self.code, locator=locator)
        super().__init__(text)
        self.locator = locator
        self.text = text
        self.code = code or self.code

    def as_xml(self):
        return render_to_string(
            [
                f"gisserver/{self.service.lower()}/{self.version}/exception.xml",
                f"gisserver/{self.version}/exception.xml",
            ],
            {"self": self},
        )

    def __html__(self):
        return self.as_xml()


class WFSException(OWSException):
    service = "WFS"


class OperationNotSupported(WFSException):
    status_code = 400
    code = "OperationNotSupported"
    text_template = "Operation is not implemented."


class MissingParameterValue(WFSException):
    status_code = 400
    code = "MissingParameterValue"
    text_template = "Missing required '{locator}' parameter."


class InvalidParameterValue(WFSException):
    status_code = 400
    code = "InvalidParameterValue"
    text_template = "Invalid value for '{locator}' parameter."


class VersionNegotiationFailed(WFSException):
    status_code = 400
    code = "VersionNegotiationFailed"
    text_template = "'ACCEPTVERSIONS' contains an invalid version number."



from urllib.parse import urljoin

from django.contrib.gis.geos import GEOSGeometry, Point
from django.template import Library
from django.utils.html import format_html

from gisserver.features import FeatureType
from gisserver.types import CRS

register = Library()


@register.filter(name="urljoin")
def urljoin_(fragment, server_url):
    return urljoin(server_url, fragment)


@register.filter
def with_feature_fields(instance, feature: FeatureType):
    return [
        (field, xs_type, getattr(instance, field)) for field, xs_type in feature.fields
    ]


@register.simple_tag
def gml_bounds(feature: FeatureType, instance, output_crs: CRS):
    """Generate the <gml:boundedBy>> for an instance."""
    bbox = feature.get_envelope(instance, output_crs)
    if bbox is None:
        return ""

    return format_html(
        """<gml:boundedBy>
        <gml:Envelope srsName="{srs_name}">
            <gml:lowerCorner>{lower}</gml:lowerCorner>
            <gml:upperCorner>{upper}</gml:upperCorner>
        </gml:Envelope>
    </gml:boundedBy>""",
        srs_name=str(output_crs),
        instance_id=instance.pk,
        lower=" ".join(map(str, bbox.lower_corner)),
        upper=" ".join(map(str, bbox.upper_corner)),
    )


@register.simple_tag(takes_context=True)
def gml_value(context, prefix: str, id, value: GEOSGeometry, output_crs: CRS):
    """Generate the GML for a geometry value."""
    forloop = context["forloop"]
    gml_field_index = forloop.get("gml_field_index", 0) + 1
    forloop["gml_field_index"] = gml_field_index

    if isinstance(value, Point):
        value.transform(output_crs.srid)
        return format_html(
            """<gml:Point gml:id="{prefix}.{id}.{seq}" srsName="{srs_name}">
            <gml:pos>{coords}</gml:pos>
          </gml:Point>""",
            prefix=prefix,
            id=id,
            seq=gml_field_index,
            srs_name=str(output_crs),
            coords=" ".join(map(str, value.coords)),
        )
    else:
        raise NotImplementedError(f"Rendering {value} is not implemented")



"""Dataclasses that expose the metadata for the GetCapabilities call."""
from __future__ import annotations

from dataclasses import dataclass, field
from math import inf
from typing import List, Optional, Tuple, Type

from django.contrib.gis.db.models import Extent, GeometryField
from django.contrib.gis.db.models.functions import Transform
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.functional import cached_property  # py3.8: functools

from gisserver.types import CRS, WGS84, BoundingBox


@dataclass
class ServiceDescription:
    """Basic metadata for an exposed GIS service."""

    title: str
    abstract: Optional[str] = None
    keywords: Optional[List[str]] = None

    provider_name: Optional[str] = None
    provider_site: Optional[str] = None
    contact_person: Optional[str] = None


@dataclass
class FeatureType:
    """Declare a feature that is exposed on the map.

    This corresponds with a single Django model.
    """

    model: Type[models.Model]
    geometry_field_name: str = None

    # WFS Metadata:
    name: str = None
    title: str = None
    abstract: str = None
    keywords: List[str] = field(default_factory=list)
    crs: CRS = None
    other_crs: List[CRS] = field(default_factory=list)
    metadata_url: Optional[str] = None

    def __post_init__(self):
        # Add some defaults
        if not self.name:
            self.name = self.model._meta.model_name
        if not self.title:
            self.title = self.model._meta.verbose_name

        self.geometry_fields = [
            f for f in self.model._meta.get_fields() if isinstance(f, GeometryField)
        ]
        self.geometry_field = self._get_geometry_field()
        default_crs = self.geometry_field.srid  # checks lookup too
        if not self.crs:
            self.crs = CRS.from_string(default_crs)

    @cached_property
    def geometry_field_names(self):
        return [f.name for f in self.geometry_fields]

    @cached_property
    def fields(self) -> List[Tuple[str, str]]:
        """Return all fields that can be queried."""
        fields = []
        for model_field in self.model._meta.get_fields():
            if isinstance(model_field, models.ForeignKey):
                # Don't let it query on the relation value yet
                field_name = model_field.attname
            else:
                field_name = model_field.name

            field_type = self.get_field_type(model_field)
            fields.append((field_name, field_type))

        return fields

    def get_field_type(self, model_field) -> str:
        """Determine the XSD field type for a Django field."""
        if isinstance(model_field, models.ForeignKey):
            # Don't let it query on the relation value yet
            return self.get_field_type(model_field.remote_field)
        elif isinstance(model_field, models.IntegerField):
            return "integer"
        elif model_field.name == self.geometry_field_name:
            return "gml:GeometryPropertyType"
        else:
            return "string"

    def _get_geometry_field(self) -> GeometryField:
        """Access the Django field"""
        if not self.geometry_fields:
            raise ImproperlyConfigured(
                f"Django model {self.model.__name__} does not have to a geometry field."
            ) from None

        if self.geometry_field_name:
            return self.model._meta.get_field(self.geometry_field_name)
        else:
            self.geometry_field_name = self.geometry_fields[0].name
            return self.geometry_fields[0]

    def get_bounding_box(self) -> Optional[BoundingBox]:
        """Returns a WGS84 BoundingBox."""
        if self.geometry_field.srid != WGS84.srid:
            geo_expression = Transform(self.geometry_field.name, WGS84.srid)
        else:
            geo_expression = self.geometry_field.name

        bbox = self.model.objects.aggregate(a=Extent(geo_expression))["a"]
        return BoundingBox(*bbox, crs=WGS84) if bbox else None

    def get_envelope(self, instance, crs: CRS) -> Optional[BoundingBox]:
        """Get the bounding box for a single instance"""
        bbox = BoundingBox(inf, inf, -inf, -inf, crs=crs)

        for model_field in self.geometry_fields:
            geometry = getattr(instance, model_field.name)
            if geometry is not None:
                bbox.extend_to_geometry(geometry)

        return bbox if bbox.lower_lat != inf else None



from environ import Env

env = Env()

SECRET_KEY = "insecure-tests-only"

DATABASES = {
    "default": env.db_url(
        "DATABASE_URL",
        default="postgres://localhost/gisserver",  # default homebrew user is superuser
        engine="django.contrib.gis.db.backends.postgis",
    )
}

INSTALLED_APPS = [
    "gisserver",
    "tests.test_gisserver",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "OPTIONS": {"loaders": ["django.template.loaders.app_directories.Loader",],},
    },
]


CSRF_COOKIE_SECURE = False
SESSION_COOKIE_SECURE = False

PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]

ROOT_URLCONF = __name__

# urls.py part:
urlpatterns = []



"""Operations that implement the WFS 2.0 spec.

Useful docs:
* http://portal.opengeospatial.org/files/?artifact_id=39967
* https://www.opengeospatial.org/standards/wfs#downloads
* http://schemas.opengis.net/wfs/2.0/
* https://mapserver.org/development/rfc/ms-rfc-105.html
* https://enonline.supermap.com/iExpress9D/API/WFS/WFS200/WFS_2.0.0_introduction.htm
"""
from __future__ import annotations

import math
import re
from urllib.parse import urlencode

import ujson
from django.contrib.gis.geos import Point
from django.http import HttpResponse

from gisserver.exceptions import InvalidParameterValue, VersionNegotiationFailed
from gisserver.features import FeatureType
from gisserver.types import CRS, BoundingBox

from .base import (
    OutputFormat,
    Parameter,
    WFSFeatureMethod,
    WFSMethod,
    UnsupportedParameter,
)

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")
RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.


class GetCapabilities(WFSMethod):
    """"This operation returns map features, and available operations this WFS server supports."""

    output_formats = [OutputFormat("text/xml")]
    xml_template_name = "get_capabilities.xml"

    def get_parameters(self):
        # Not calling super, as this differs slightly (e.g. no output formats)
        return [
            Parameter(
                "service",
                required=True,
                in_capabilities=True,
                allowed_values=list(self.view.accept_operations.keys()),
            ),
            Parameter(
                # Version parameter can still be sent,
                # but not mandatory for GetCapabilities
                "version",
                allowed_values=self.view.accept_versions,
            ),
            Parameter(
                "AcceptVersions",
                in_capabilities=True,
                allowed_values=self.view.accept_versions,
                parser=self._parse_accept_versions,
            ),
            Parameter(
                "AcceptFormats",
                in_capabilities=True,
                allowed_values=self.output_formats,
                parser=self._parse_output_format,
                default=self.output_formats[0],
            ),
        ] + self.parameters

    def _parse_accept_versions(self, accept_versions) -> str:
        """Special parsing for the ACCEPTVERSIONS parameter."""
        if "VERSION" in self.view.KVP:
            # Even GetCapabilities can still receive a VERSION argument to fixate it.
            raise InvalidParameterValue(
                "AcceptVersions", "Can't provide both ACCEPTVERSIONS and VERSION"
            )

        matched_versions = set(accept_versions.split(",")).intersection(
            self.view.accept_versions
        )
        if not matched_versions:
            allowed = ", ".join(self.view.accept_versions)
            raise VersionNegotiationFailed(
                "acceptversions",
                f"'{accept_versions}' does not contain supported versions, "
                f"supported are: {allowed}.",
            )

        # Take the highest version (mapserver returns first matching)
        requested_version = sorted(matched_versions, reverse=True)[0]

        # Make sure the views+exceptions continue to operate in this version
        self.view.set_version(requested_version)

        return requested_version

    def __call__(self, **params):
        """GetCapabilities only supports XML output"""
        context = self.get_context_data(**params)
        return self.render_xml(context, **params)

    def get_context_data(self, **params):
        view = self.view

        # The 'service' is not reaad from 'params' to avoid dependency on get_parameters()
        service = view.KVP["SERVICE"]  # is WFS
        service_operations = view.accept_operations[service]
        feature_output_formats = service_operations["GetFeature"].output_formats

        return {
            "service_description": view.get_service_description(service),
            "accept_operations": {
                name: operation(view) for name, operation in service_operations.items()
            },
            "service_constraints": self.view.service_constraints,
            "filter_capabilities": self.view.filter_capabilities,
            "accept_versions": self.view.accept_versions,
            "feature_types": self.view.get_feature_types(),
            "feature_output_formats": feature_output_formats,
            "default_max_features": self.view.max_page_size,
        }


class DescribeFeatureType(WFSFeatureMethod):
    """This returns an XML Schema for the provided objects.
    Each feature is exposed as an XSD definition with it's fields.
    """

    output_formats = [
        OutputFormat("XMLSCHEMA"),
        # OutputFormat("text/xml", subtype="gml/3.1.1"),
    ]
    xml_template_name = "describe_feature_type.xml"
    xml_content_type = "application/gml+xml; version=3.2"  # mandatory for WFS

    def get_context_data(self, typeNames, **params):
        return {"feature_types": typeNames}


class GetFeature(WFSFeatureMethod):
    """This returns the feature details; the individual records based on a query.
    Various query parameters allow limiting the data.
    """

    parameters = [
        Parameter(
            "resultType",
            in_capabilities=True,
            parser=lambda x: x.upper(),
            allowed_values=["RESULTS", "HITS"],
            default="RESULTS",
        ),
        Parameter("srsName", parser=CRS.from_string),
        Parameter("bbox", parser=BoundingBox.from_string),
        Parameter("startIndex", parser=int, default=0),
        Parameter("count", parser=int),  # was called maxFeatures in WFS 1.x
        # Not implemented:
        UnsupportedParameter("filter"),
        UnsupportedParameter("sortBy"),
        UnsupportedParameter("resourceID"),  # query on ID (called featureID in wfs 1.x)
        UnsupportedParameter("propertyName"),  # which fields to return
        UnsupportedParameter("resolve"),  # subresource settings
        UnsupportedParameter("resolveDepth"),
        UnsupportedParameter("resolveTimeout"),
        UnsupportedParameter("namespaces"),  # define output namespaces
        UnsupportedParameter("aliases"),
        UnsupportedParameter("filter_language"),
    ]
    output_formats = [
        OutputFormat("text/xml", subtype="gml/3.1.1"),
        # OutputFormat("gml"),
        OutputFormat("application/json", subtype="geojson", charset="utf-8"),
        # OutputFormat("text/csv"),
        # OutputFormat("shapezip"),
        # OutputFormat("application/zip"),
    ]
    xml_template_name = "get_feature.xml"

    def get_context_data(self, resultType, **params):
        if resultType == "HITS":
            return self.get_context_data_hits(**params)
        elif resultType == "RESULTS":
            return self.get_context_data_results(**params)
        else:
            raise NotImplementedError()

    def get_context_data_hits(self, srsName, **params):
        """Return the context data for the "HITS" type."""
        querysets = self.get_querysets(**params)
        number_matched = sum([qs.count() for qs in querysets])
        return {
            "output_crs": srsName or params["typeNames"][0].crs,
            "feature_collections": [],
            "number_matched": number_matched,
        }

    def get_context_data_results(self, srsName, **params):
        """Return the context data for the "RESULTS" type."""
        querysets = self.get_querysets(**params)

        # Parse pagination settings
        max_page_size = self.view.max_page_size
        start = max(0, params["startIndex"])
        page_size = min(max_page_size, params["count"] or max_page_size)
        stop = start + page_size

        # The querysets are not executed until the very end.
        feature_collections = [
            (qs.feature, qs[start:stop].iterator(), qs.count()) for qs in querysets
        ]
        number_matched = sum(item[2] for item in feature_collections)

        previous = next = None
        if start > 0:
            previous = self._replace_url_params(STARTINDEX=max(0, start - page_size))
        if stop < number_matched:  # TODO: fix for returning multiple typeNames
            next = self._replace_url_params(STARTINDEX=start + page_size)

        return {
            "output_crs": srsName or params["typeNames"][0].crs,
            "feature_collections": feature_collections,
            "number_matched": number_matched,
            "next": next,
            "previous": previous,
        }

    def get_querysets(self, typeNames, **params):
        """Generate querysets for all requested data."""
        return [
            self.filter_queryset(feature, feature.model.objects.all(), **params)
            for feature in typeNames
        ]

    def filter_queryset(self, feature: FeatureType, queryset, bbox, **params):
        """Apply the filters to a single feature type."""
        filters = {}

        # Allow filtering within a bounding box
        if bbox:
            filters[f"{feature.geometry_field_name}__within"] = bbox.as_polygon()

        # TODO: other parameters to support:
        # filter=<fes:...>
        # sortBy=attr+A / +D
        # resourceid=app:Type/gml:name (was featureID in wfs 1.x)
        # propertyName=app:Type/app:field1,app:Type/app:field2
        for name in ("filter", "sortBy", "resourceID", "propertyName"):
            if params[name]:
                raise InvalidParameterValue(name, f"{name} is not supported yet")

        if filters:
            queryset = queryset.filter(**filters)

        # Attach extra property to keep meta-dataa in place
        queryset.feature = feature
        return queryset

    def render_xml(self, context, **params):
        """Improve the context for XML output."""
        # It's not possible the stream the queryset results with the XML format,
        # as the first tag needs to describe the number of results.
        feature_collections = [
            (feature, list(qs), matched)
            for feature, qs, matched in context["feature_collections"]
        ]

        # Determine bounding box of all items
        bounding_box = BoundingBox(math.inf, math.inf, -math.inf, -math.inf)
        for feature, qs, matched in feature_collections:
            for instance in qs:
                geomery_value = getattr(instance, feature.geometry_field_name)
                if geomery_value is None:
                    continue

                if isinstance(geomery_value, Point):
                    bounding_box.extend_to(
                        geomery_value.x,
                        geomery_value.y,
                        geomery_value.x,
                        geomery_value.y,
                    )
                else:
                    raise NotImplementedError(
                        f"Rendering {geomery_value} is not implemented"
                    )

        context["feature_collections"] = feature_collections
        context["number_returned"] = sum(len(qs) for _, qs, _ in feature_collections)
        context["bounding_box"] = bounding_box
        context["xsd_typenames"] = self.view.KVP["TYPENAMES"]
        return super().render_xml(context)

    def render_geojson(self, context, **params):
        # TODO: write JSON as stream instead.

        # Flatten the results, they are not grouped in a second FeatureCollection
        features = []
        for feature, qs, number_matched in context["feature_collections"]:
            for instance in qs:
                geo_value = getattr(instance, feature.geometry_field_name)
                features.append(
                    {
                        "type": "Feature",
                        "id": f"{feature.name}.{instance.pk}",  # foreign member (allowed by spec)
                        "geometry_name": str(
                            instance
                        ),  # foreign member (allowed by spec)
                        # bbox: [x, y, x, y]
                        "geometry": (
                            {
                                # using geom_class instead of geom_type for CamelCasing
                                "type": feature.geometry_field.geom_class.__name__,
                                "coordinates": geo_value.coords,
                            }
                            if geo_value is not None
                            else None
                        ),
                        "properties": {
                            name: getattr(instance, name)
                            for name, _ in feature.fields
                            if name not in feature.geometry_field_names
                        },
                    }
                )

        crs = context["output_crs"]
        context = {
            "type": "FeatureCollection",
            "totalFeatures": context["number_matched"],
            "crs": {"type": "name", "properties": {"name": str(crs)}},
            "features": features,
        }

        return HttpResponse(
            ujson.dumps(context), content_type="application/json; charset=utf-8"
        )

    def _replace_url_params(self, **updates) -> str:
        """Replace a query parameter in the URL"""
        new_params = self.view.KVP.copy()
        new_params.update(updates)
        return f"{self.view.server_url}?{urlencode(new_params)}"



from gisserver.types import CRS, WGS84


class TestCRS:
    def test_from_string(self):
        assert CRS.from_string(4326).urn == "urn:ogc:def:crs:EPSG::4326"
        assert CRS.from_string("EPSG:4326").urn == "urn:ogc:def:crs:EPSG::4326"

    def test_crs84_inequal(self):
        """Prove that using strings from other vendors is also parsed.
        """
        crs84 = CRS.from_string("urn:ogc:def:crs:OGC:1.3:CRS84")
        assert crs84.srid == WGS84.srid

        # EPSG:4326 specifies coordinates in lat/long order and CRS:84 in long/lat order
        assert crs84 != WGS84

    def test_crs_empty_version(self):
        """Prove that empty versions are properly re-encoded as empty string"""
        assert (
            CRS.from_string("urn:ogc:def:crs:EPSG::28992").urn
            == "urn:ogc:def:crs:EPSG::28992"
        )



"""The base protocol to implement an operation.

All operations extend from an WFSMethod class.
This defines the parameters and output formats of the method.
This introspection data is also parsed by the GetCapabilities call.
"""
from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponse
from django.template.loader import render_to_string

from gisserver.exceptions import InvalidParameterValue, MissingParameterValue
from gisserver.features import FeatureType

NoneType = type(None)

RE_SAFE_FILENAME = re.compile(r"\A[A-Za-z0-9]+[A-Za-z0-9.]*")  # no dot at the start.


@dataclass(frozen=True)
class Parameter:
    """The definition of an parameter for an WFS method.

    These parameters should be defined in the
    WFSMethod parameters attribute / get_parameters().

    These fields is used to parse the incoming request.
    The GetCapabilities method also reads the metadata of this value.
    """

    #: Name of the parameter (using the XML casing style)
    name: str

    #: Whether the parameter is required
    required: bool = False

    #: Optional parser callback to convert the parameter into a Python type
    parser: Callable[[str], Any] = None

    #: Whether to list this parameter in GetCapabilities
    in_capabilities: bool = False

    #: List of allowed values (also shown in GetCapabilities)
    allowed_values: Union[list, tuple, set, NoneType] = None

    #: Default value if it's not given
    default: Any = None

    #: Overridable dict for error messages
    error_messages: dict = field(default_factory=dict)

    def value_from_query(self, KVP: dict):  # noqa: C901
        """Parse a request variable using the type definition.

        This uses the dataclass settings to parse the incoming request value.
        """
        # URL-based key-value-pair parameters use uppercase.
        kvp_name = self.name.upper()
        value = KVP.get(kvp_name)

        # Check required field settings, both empty and missing value are treated the same.
        if not value:
            if self.required:
                raise MissingParameterValue(self.name, f"Empty {kvp_name} parameter")
            else:
                return self.default

        # Allow conversion into a python object
        if self.parser is not None:
            try:
                value = self.parser(value)
            except (TypeError, ValueError) as e:
                raise InvalidParameterValue(
                    self.name, f"Unable to parse {kvp_name} argument: {e}"
                ) from None

        # Validate against value choices
        self.validate_value(value)

        return value

    def validate_value(self, value):
        """Validate the parsed value.
        This method can be overwritten by a subclass if needed.
        """
        if self.allowed_values is not None and value not in self.allowed_values:
            msg = self.error_messages.get(
                "invalid", "Invalid value for {name}: {value}"
            )
            raise InvalidParameterValue(
                self.name, msg.format(name=self.name, value=value)
            )


class UnsupportedParameter(Parameter):
    def value_from_query(self, KVP: dict):
        kvp_name = self.name.upper()
        if kvp_name in KVP:
            raise InvalidParameterValue(
                self.name, f"Support for {self.name} is not implemented!"
            )
        return None


class OutputFormat:
    """Declare an output format for the method.

    These formats should be used in the ``output_formats`` section of the WFSMethod.
    """

    def __init__(self, content_type, **extra):
        self.content_type = content_type
        self.extra = extra
        self.subtype = self.extra.get("subtype")

    def matches(self, value):
        """Test whether the 'value' is matched by this object."""
        return self.content_type == value or self.subtype == value

    def __str__(self):
        extra = "".join(f"; {name}={value}" for name, value in self.extra.items())
        return f"{self.content_type}{extra}"


class WFSMethod:
    """Basic interface to implement an WFS method.

    Each operation in this GIS-server extends from this base class. This class
    also exposes all requires metadata for the GetCapabilities request.
    """

    do_not_call_in_templates = True  # avoid __call__ execution in Django templates

    #: List the suported parameters for this method, extended by get_parameters()
    parameters: List[Parameter] = []

    #: List the supported output formats for this method.
    output_formats: List[OutputFormat] = []

    #: Default template to use for rendering
    xml_template_name = None

    #: Default content-type for render_xml()
    xml_content_type = "text/xml; charset=utf-8"

    def __init__(self, view: OGCView):
        self.view = view

    def get_parameters(self):
        """Dynamically return the supported parameters for this method."""
        parameters = [
            # Always add SERVICE and VERSION as required parameters
            Parameter(
                "service",
                required=True,
                in_capabilities=True,
                allowed_values=list(self.view.accept_operations.keys()),
                error_messages={"invalid": "Unsupported service type: {value}."},
            ),
            Parameter(
                "version",
                required=True,  # Mandatory except for GetCapabilities
                allowed_values=self.view.accept_versions,
                error_messages={
                    "invalid": "WFS Server does not support VERSION {value}.",
                },
            ),
        ] + self.parameters

        if self.output_formats:
            parameters += [
                Parameter(
                    "outputFormat",
                    in_capabilities=True,
                    allowed_values=self.output_formats,
                    parser=self._parse_output_format,
                    default=self.output_formats[0],
                )
            ]

        return parameters

    def _parse_output_format(self, value) -> OutputFormat:
        """Select the proper OutputFormat object based on the input value"""
        try:
            return next(o for o in self.output_formats if o.matches(value))
        except StopIteration:
            raise InvalidParameterValue(
                "outputformat",
                f"'{value}' is not a permitted output format for this operation.",
            ) from None

    def parse_request(self, KVP: dict) -> Dict[str, Any]:
        """Parse the parameters of the request"""
        param_values = {
            param.name: param.value_from_query(KVP) for param in self.get_parameters()
        }

        # Update version if requested.
        # This is stored on the view, so exceptions also use it.
        if param_values.get("version"):
            self.view.set_version(param_values["version"])

        return param_values

    def __call__(self, **params):
        """Default call implementation: render an XML template."""
        context = self.get_context_data(**params)
        output_format: OutputFormat = params["outputFormat"]
        if output_format.subtype == "geojson":
            return self.render_geojson(context, **params)
        else:
            return self.render_xml(context, **params)

    def get_context_data(self, **params):
        """Collect all arguments to use for rendering the XML template"""
        return {}

    def render_geojson(self, context, **params):
        """Shortcut to render JSON"""
        raise NotImplementedError()

    def render_xml(self, context, **params):
        """Shortcut to render XML"""
        return HttpResponse(
            render_to_string(
                self._get_xml_template_name(),
                context={
                    "view": self.view,
                    "app_xml_namespace": self.view.xml_namespace,
                    "server_url": self.view.server_url,
                    **context,
                },
            ),
            content_type=self.xml_content_type,
        )

    def _get_xml_template_name(self):
        """Generate the XML template name for this operation, and check it's file pattern"""
        service = self.view.KVP["SERVICE"].lower()
        template_name = (
            f"gisserver/{service}/{self.view.version}/{self.xml_template_name}"
        )

        # Since 'service' and 'version' are based on external input,
        # these values are double checked again to avoid remove file inclusion.
        if not RE_SAFE_FILENAME.match(service) or not RE_SAFE_FILENAME.match(
            self.view.version
        ):
            raise SuspiciousOperation(
                f"Refusing to render template name {template_name}"
            )

        return template_name


class WFSFeatureMethod(WFSMethod):
    """A base method that also resolved the TYPENAMES parameter."""

    def get_parameters(self):
        return super().get_parameters() + [
            # QGis sends both TYPENAME (wfs 1.x) and TYPENAMES (wfs 2.0) for DescribeFeatureType
            Parameter("typeNames", required=True, parser=self._parse_type_names)
        ]

    def _parse_type_names(self, type_names) -> List[FeatureType]:
        """Find the requested feature types by name"""
        self.feature_types = self.view.get_feature_types()

        type_names = type_names.split(",")
        features_by_name = {feature.name: feature for feature in self.feature_types}
        features = []
        for name in type_names:
            try:
                feature = features_by_name[name]
            except KeyError:
                raise InvalidParameterValue(
                    "typename",
                    f"Typename '{name}' doesn't exist in this server. "
                    f"Please check the capabilities and reformulate your request.",
                ) from None

            features.append(feature)
        return features


if TYPE_CHECKING:
    from gisserver.views import OGCView



import pytest
from django.urls import path
from lxml import etree

from gisserver.features import FeatureType, ServiceDescription
from gisserver.types import WGS84, CRS
from gisserver.views import WFSView

from .models import Restaurant
from .utils import WFS_20_XSD, assert_xml_equal, validate_xsd

WFS_NS = "http://www.opengis.net/wfs/2.0"
OWS_NS = "http://www.opengis.net/ows/1.1"
XLINK_NS = "http://www.w3.org/1999/xlink"
NAMESPACES = {
    "app": "http://example.org/gisserver",
    "gml": "http://www.opengis.net/gml/3.2",
    "ows": "http://www.opengis.net/ows/1.1",
    "wfs": "http://www.opengis.net/wfs/2.0",
    "xsd": "http://www.w3.org/2001/XMLSchema",
}

RD_NEW = CRS.from_string("urn:ogc:def:crs:EPSG::28992")


class PlacesWFSView(WFSView):
    """An simple view that uses the WFSView against our test model."""

    xml_namespace = "http://example.org/gisserver"
    service_description = ServiceDescription(
        # While not tested directly, this is still validated against the XSD
        title="Places",
        abstract="Unittesting",
        keywords=["django-gisserver"],
        provider_name="Django",
        provider_site="https://www.example.com/",
        contact_person="django-gisserver",
    )
    feature_types = [
        FeatureType(
            Restaurant,
            keywords=["unittest"],
            other_crs=[RD_NEW],
            metadata_url="/feature/restaurants/",
        ),
    ]


urlpatterns = [
    path("v1/wfs/", PlacesWFSView.as_view(), name="wfs-view"),
]

pytestmark = [pytest.mark.urls(__name__)]  # enable for all tests in this file


@pytest.mark.django_db
class TestGetCapabilities:
    """All tests for the GetCapabilities method."""

    def test_get(self, client, restaurant):
        """Prove that the happy flow works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetCapabilities&ACCEPTVERSIONS=2.0.0"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "<ows:OperationsMetadata>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["version"] == "2.0.0"

        # Check exposed allowed vesions
        allowed_values = xml_doc.xpath(
            "ows:OperationsMetadata/ows:Operation[@name='GetCapabilities']"
            "/ows:Parameter[@name='AcceptVersions']/ows:AllowedValues",
            namespaces=NAMESPACES,
        )[0]
        versions = [el.text for el in allowed_values.findall("ows:Value", NAMESPACES)]
        assert versions == ["2.0.0"]

        # Check exposed FeatureTypeList
        feature_type_list = xml_doc.find("wfs:FeatureTypeList", NAMESPACES)

        # The box should be within WGS84 limits, otherwise gis tools can't process the service.
        wgs84_bbox = feature_type_list.find(
            "wfs:FeatureType/ows:WGS84BoundingBox", NAMESPACES
        )
        lower = wgs84_bbox.find("ows:LowerCorner", NAMESPACES).text.split(" ")
        upper = wgs84_bbox.find("ows:UpperCorner", NAMESPACES).text.split(" ")
        coords = list(map(float, lower + upper))
        assert coords[0] >= -180
        assert coords[1] >= -90
        assert coords[2] <= 180
        assert coords[2] <= 90

        assert_xml_equal(
            etree.tostring(feature_type_list, inclusive_ns_prefixes=True).decode(),
            f"""<FeatureTypeList xmlns="{WFS_NS}" xmlns:ows="{OWS_NS}" xmlns:xlink="{XLINK_NS}">
      <FeatureType>
        <Name>restaurant</Name>
        <Title>restaurant</Title>
        <ows:Keywords>
          <ows:Keyword>unittest</ows:Keyword>
        </ows:Keywords>
        <DefaultCRS>urn:ogc:def:crs:EPSG::4326</DefaultCRS>
        <OtherCRS>urn:ogc:def:crs:EPSG::28992</OtherCRS>
        <OutputFormats>
          <Format>text/xml; subtype=gml/3.1.1</Format>
          <Format>application/json; subtype=geojson; charset=utf-8</Format>
        </OutputFormats>
        <ows:WGS84BoundingBox dimensions="2">
          <ows:LowerCorner>4.90876101285122 52.3631712637357</ows:LowerCorner>
          <ows:UpperCorner>4.90876101285122 52.3631712637357</ows:UpperCorner>
        </ows:WGS84BoundingBox>
        <MetadataURL xlink:href="http://testserver/v1/wfs/" />
      </FeatureType>
    </FeatureTypeList>""",
        )

    def test_missing_parameters(self, client):
        """Prove that missing arguments are handled"""
        response = client.get("/v1/wfs/?SERVICE=WFS")
        content = response.content.decode()
        assert response.status_code == 400, content
        assert response["content-type"] == "text/xml; charset=utf-8", content

        assert_xml_equal(
            response.content,
            """<ows:ExceptionReport version="2.0.0"
 xmlns:ows="http://www.opengis.net/ows/1.1"
 xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 xml:lang="en-US"
 xsi:schemaLocation="http://www.opengis.net/ows/1.1 http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd">

  <ows:Exception exceptionCode="MissingParameterValue" locator="request">
    <ows:ExceptionText>Missing required &#x27;request&#x27; parameter.</ows:ExceptionText>
  </ows:Exception>
</ows:ExceptionReport>""",  # noqa: E501
        )

        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["version"] == "2.0.0"
        exception = xml_doc.find("ows:Exception", NAMESPACES)
        assert exception.attrib["exceptionCode"] == "MissingParameterValue"

    def test_version_negotiation(self, client):
        """Prove that version negotiation still returns 2.0.0"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetCapabilities&ACCEPTVERSIONS=1.0.0,2.0.0"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content

        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["version"] == "2.0.0"

    def test_get_invalid_version(self, client):
        """Prove that version negotiation works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetCapabilities&ACCEPTVERSIONS=1.5.0"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 400, content

        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        exception = xml_doc.find("ows:Exception", NAMESPACES)
        assert exception.attrib["exceptionCode"] == "VersionNegotiationFailed"


class TestDescribeFeatureType:
    """All tests for the DescribeFeatureType method."""

    def test_get(self, client):
        """Prove that the happy flow works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=DescribeFeatureType&VERSION=2.0.0&TYPENAMES=restaurant"
        )
        content = response.content.decode()
        assert response["content-type"] == "application/gml+xml; version=3.2", content
        assert response.status_code == 200, content
        assert "gml:GeometryPropertyType" in content

        # The response is an XSD itself.
        # Only validate it's XML structure
        xml_doc: etree._Element = etree.fromstring(response.content)
        assert xml_doc.tag == "{http://www.w3.org/2001/XMLSchema}schema"
        elements = xml_doc.findall(
            "xsd:complexType/xsd:complexContent/xsd:extension/xsd:sequence/xsd:element",
            namespaces=NAMESPACES,
        )
        field_names = [el.attrib["name"] for el in elements]
        assert field_names == ["id", "name", "location"]

        assert_xml_equal(
            response.content,
            """<schema
   targetNamespace="http://example.org/gisserver"
   xmlns:app="http://example.org/gisserver"
   xmlns:xsd="http://www.w3.org/2001/XMLSchema"
   xmlns="http://www.w3.org/2001/XMLSchema"
   xmlns:gml="http://www.opengis.net/gml/3.2"
   elementFormDefault="qualified" version="0.1" >

  <import namespace="http://www.opengis.net/gml/3.2" schemaLocation="http://schemas.opengis.net/gml/3.2.1/gml.xsd" />
  <element name="restaurant" type="app:restaurantType" substitutionGroup="gml:AbstractFeature" />

  <complexType name="restaurantType">
    <complexContent>
      <extension base="gml:AbstractFeatureType">
        <sequence>
          <element name="id" minOccurs="0" type="integer"/>
          <element name="name" minOccurs="0" type="string"/>
          <element name="location" type="gml:GeometryPropertyType" minOccurs="0" maxOccurs="1"/>
        </sequence>
      </extension>
    </complexContent>
  </complexType>
</schema>""",  # noqa: E501
        )

    def test_empty_typenames(self, client):
        """Prove that missing arguments are handled"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=DescribeFeatureType&VERSION=2.0.0&TYPENAMES="
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 400, content
        assert "</ows:Exception>" in content

        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["version"] == "2.0.0"
        exception = xml_doc.find("ows:Exception", NAMESPACES)
        assert exception.attrib["exceptionCode"] == "MissingParameterValue"


@pytest.mark.django_db
class TestGetFeature:
    """All tests for the GetFeature method.
    The methods need to have at least one datatype, otherwise not all content is rendered.
    """

    def test_get(self, client, restaurant):
        """Prove that the happy flow works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "</wfs:FeatureCollection>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["numberMatched"] == "1"
        assert xml_doc.attrib["numberReturned"] == "1"

        # See whether our feature is rendered
        feature = xml_doc.find("wfs:member/app:restaurant", namespaces=NAMESPACES)
        assert feature is not None
        assert feature.find("app:name", namespaces=NAMESPACES).text == restaurant.name
        timestamp = xml_doc.attrib["timeStamp"]

        assert_xml_equal(
            response.content,
            f"""<wfs:FeatureCollection
   xmlns:app="http://example.org/gisserver"
   xmlns:gml="http://www.opengis.net/gml/3.2"
   xmlns:wfs="http://www.opengis.net/wfs/2.0"
   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
   xsi:schemaLocation="http://example.org/gisserver http://testserver/v1/wfs/?SERVICE=WFS&amp;VERSION=2.0.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAMES=restaurant http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd"
   timeStamp="{timestamp}" numberMatched="1" numberReturned="1">

    <wfs:member>
      <app:restaurant gml:id="restaurant.{restaurant.id}">
        <gml:boundedBy>
            <gml:Envelope srsName="urn:ogc:def:crs:EPSG::4326">
                <gml:lowerCorner>4.908761012851219 52.363171263735715</gml:lowerCorner>
                <gml:upperCorner>4.908761012851219 52.363171263735715</gml:upperCorner>
            </gml:Envelope>
        </gml:boundedBy>
        <app:id>{restaurant.id}</app:id>
        <app:name>Café Noir</app:name>
        <app:location>
          <gml:Point gml:id="restaurant.{restaurant.id}.1" srsName="urn:ogc:def:crs:EPSG::4326">
            <gml:pos>4.908761012851219 52.363171263735715</gml:pos>
          </gml:Point>
        </app:location>
      </app:restaurant>
    </wfs:member>
</wfs:FeatureCollection>""",  # noqa: E501
        )

    def test_get_empty_geometry(self, client):
        """Prove that the empty geometry values don't crash the rendering."""
        restaurant = Restaurant.objects.create(name="Empty")
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "</wfs:FeatureCollection>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        feature = xml_doc.find("wfs:member/app:restaurant", namespaces=NAMESPACES)
        assert feature.find("app:location", namespaces=NAMESPACES).text is None
        timestamp = xml_doc.attrib["timeStamp"]

        assert_xml_equal(
            response.content,
            f"""<wfs:FeatureCollection
       xmlns:app="http://example.org/gisserver"
       xmlns:gml="http://www.opengis.net/gml/3.2"
       xmlns:wfs="http://www.opengis.net/wfs/2.0"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://example.org/gisserver http://testserver/v1/wfs/?SERVICE=WFS&amp;VERSION=2.0.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAMES=restaurant http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd"
       timeStamp="{timestamp}" numberMatched="1" numberReturned="1">

        <wfs:member>
          <app:restaurant gml:id="restaurant.{restaurant.id}">
            <app:id>{restaurant.id}</app:id>
            <app:name>Empty</app:name>
            <app:location />
          </app:restaurant>
        </wfs:member>
    </wfs:FeatureCollection>""",  # noqa: E501
        )

    def test_get_srs_name(self, client, restaurant):
        """Prove that specifying SRSNAME works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
            "&SRSNAME=urn:ogc:def:crs:EPSG::28992"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "</wfs:FeatureCollection>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["numberMatched"] == "1"
        assert xml_doc.attrib["numberReturned"] == "1"

        # Prove that the output is now rendered in EPSG:28992
        feature = xml_doc.find("wfs:member/app:restaurant", namespaces=NAMESPACES)
        geometry = feature.find("app:location/gml:Point", namespaces=NAMESPACES)
        assert geometry.attrib["srsName"] == "urn:ogc:def:crs:EPSG::28992"

        timestamp = xml_doc.attrib["timeStamp"]
        assert_xml_equal(
            response.content,
            f"""<wfs:FeatureCollection
       xmlns:app="http://example.org/gisserver"
       xmlns:gml="http://www.opengis.net/gml/3.2"
       xmlns:wfs="http://www.opengis.net/wfs/2.0"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://example.org/gisserver http://testserver/v1/wfs/?SERVICE=WFS&amp;VERSION=2.0.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAMES=restaurant http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd"
       timeStamp="{timestamp}" numberMatched="1" numberReturned="1">

        <wfs:member>
          <app:restaurant gml:id="restaurant.{restaurant.id}">
            <gml:boundedBy>
                 <gml:Envelope srsName="urn:ogc:def:crs:EPSG::28992">
                <gml:lowerCorner>122411.00000717948 486250.0005178676</gml:lowerCorner>
                <gml:upperCorner>122411.00000717948 486250.0005178676</gml:upperCorner>
            </gml:Envelope>
            </gml:boundedBy>
            <app:id>{restaurant.id}</app:id>
            <app:name>Café Noir</app:name>
            <app:location>
              <gml:Point gml:id="restaurant.{restaurant.id}.1" srsName="urn:ogc:def:crs:EPSG::28992">
                <gml:pos>122411.00000717948 486250.0005178676</gml:pos>
              </gml:Point>
            </app:location>
          </app:restaurant>
        </wfs:member>
    </wfs:FeatureCollection>""",  # noqa: E501
        )

    def test_get_bbox(self, client, restaurant):
        """Prove that that parsing RESULTTYPE=hits works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
            "&BBOX=122400,486200,122500,486300,urn:ogc:def:crs:EPSG::28992"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "</wfs:FeatureCollection>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["numberMatched"] == "1"
        assert xml_doc.attrib["numberReturned"] == "1"

        # Prove that the output is still rendered in WGS84
        feature = xml_doc.find("wfs:member/app:restaurant", namespaces=NAMESPACES)
        geometry = feature.find("app:location/gml:Point", namespaces=NAMESPACES)
        assert geometry.attrib["srsName"] == WGS84.urn

        # Also prove that using a different BBOX gives empty results
        response2 = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
            "&BBOX=100,100,200,200,urn:ogc:def:crs:EPSG::28992"
        )
        xml_doc = validate_xsd(response2.content, WFS_20_XSD)
        assert xml_doc.attrib["numberMatched"] == "0"
        assert xml_doc.attrib["numberReturned"] == "0"

    def test_get_hits(self, client, restaurant):
        """Prove that that parsing RESULTTYPE=hits works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
            "&RESULTTYPE=hits"
        )
        content = response.content.decode()
        assert response["content-type"] == "text/xml; charset=utf-8", content
        assert response.status_code == 200, content
        assert "</wfs:FeatureCollection>" in content

        # Validate against the WFS 2.0 XSD
        xml_doc = validate_xsd(response.content, WFS_20_XSD)
        assert xml_doc.attrib["numberMatched"] == "1"
        assert xml_doc.attrib["numberReturned"] == "0"
        assert not xml_doc.getchildren()  # should not have children!
        timestamp = xml_doc.attrib["timeStamp"]

        assert_xml_equal(
            response.content,
            f"""<wfs:FeatureCollection
   xmlns:app="http://example.org/gisserver"
   xmlns:gml="http://www.opengis.net/gml/3.2"
   xmlns:wfs="http://www.opengis.net/wfs/2.0"
   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
   xsi:schemaLocation="http://example.org/gisserver http://testserver/v1/wfs/?SERVICE=WFS&amp;VERSION=2.0.0&amp;REQUEST=DescribeFeatureType&amp;TYPENAMES=restaurant http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd"
   timeStamp="{timestamp}" numberMatched="1" numberReturned="0">
</wfs:FeatureCollection>""",  # noqa: E501
        )

    def test_get_geojson(self, client, restaurant):
        """Prove that the geojson export works"""
        response = client.get(
            "/v1/wfs/?SERVICE=WFS&REQUEST=GetFeature&VERSION=2.0.0&TYPENAMES=restaurant"
            "&outputformat=geojson"
        )
        assert response["content-type"] == "application/json; charset=utf-8"
        content = response.json()
        assert response.status_code == 200, content
        assert content == {
            "type": "FeatureCollection",
            "totalFeatures": 1,
            "crs": {
                "type": "name",
                "properties": {"name": "urn:ogc:def:crs:EPSG::4326"},
            },
            "features": [
                {
                    "type": "Feature",
                    "id": f"restaurant.{restaurant.id}",
                    "geometry_name": "Café Noir",
                    "geometry": {
                        "type": "Point",
                        "coordinates": [4.9087610129, 52.3631712637],
                    },
                    "properties": {"id": restaurant.id, "name": "Café Noir"},
                }
            ],
        }



"""Internal data types

This exposes helper classes to parse GEO data types.
"""
from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Optional, Union

from django.contrib.gis.geos import GEOSGeometry, Point, Polygon

CRS_URN_REGEX = re.compile(
    r"^urn:(?P<domain>[a-z]+)"
    r":def:crs:(?P<authority>[a-z]+)"
    r":(?P<version>[0-9]+\.[0-9]+(\.[0-9]+)?)?"
    r":(?P<id>[0-9]+|crs84)"
    r"$",
    re.IGNORECASE,
)

__all__ = [
    "BoundingBox",
    "CRS",
    "WGS84",
]


@dataclass
class BoundingBox:
    """A bounding box that describes the extent of a map layer"""

    lower_lon: float
    lower_lat: float
    upper_lon: float
    upper_lat: float
    crs: Optional[CRS] = None

    @classmethod
    def from_string(cls, bbox):
        """Parse the bounding box from an input string.

        It can either be 4 coordinates, or 4 coordinates with a special reference system.
        """
        bbox = bbox.split(",")
        if not (4 <= len(bbox) <= 5):
            raise ValueError(
                f"Input does not contain bounding box, "
                f"expected 4 or 5 values, not {bbox}."
            )
        return cls(
            float(bbox[0]),
            float(bbox[1]),
            float(bbox[2]),
            float(bbox[3]),
            CRS.from_string(bbox[4]) if len(bbox) == 5 else None,
        )

    @property
    def lower_corner(self):
        return [self.lower_lon, self.lower_lat]

    @property
    def upper_corner(self):
        return [self.upper_lon, self.upper_lat]

    def __repr__(self):
        return (
            "BoundingBox("
            f"{self.lower_lon}, {self.lower_lat}, {self.upper_lon}, {self.upper_lat}"
            ")"
        )

    def extend_to(
        self, lower_lon: float, lower_lat: float, upper_lon: float, upper_lat: float
    ):
        """Expand the bounding box in-place"""
        self.lower_lon = min(self.lower_lon, lower_lon)
        self.lower_lat = min(self.lower_lat, lower_lat)
        self.upper_lon = max(self.upper_lon, upper_lon)
        self.upper_lat = max(self.upper_lat, upper_lat)

    def extend_to_geometry(self, geometry: GEOSGeometry):
        """Extend this bounding box with the coordinates of a given geometry."""
        if geometry.crs.srid != self.crs.srid:
            geometry = geometry.transform(self.crs.srid, clone=True)

        if isinstance(geometry, Point):
            self.extend_to(geometry.x, geometry.y, geometry.x, geometry.y)
        else:
            raise NotImplementedError(f"Processing {geometry} is not implemented")

    def __add__(self, other):
        """Combine both extents into a larger box."""
        if isinstance(other, BoundingBox):
            if other.crs != self.crs:
                raise ValueError(
                    "Can't combine instances with different spatial reference systems"
                )
            return BoundingBox(
                min(self.lower_lon, other.lower_lon),
                min(self.lower_lat, other.lower_lat),
                max(self.upper_lon, other.upper_lon),
                max(self.upper_lat, other.upper_lat),
            )
        else:
            return NotImplemented

    def as_polygon(self) -> Polygon:
        """Convert the value into a GEOS polygon."""
        polygon = Polygon.from_bbox(
            (self.lower_lon, self.lower_lat, self.upper_lon, self.upper_lat)
        )
        if self.crs is not None:
            polygon.srid = self.crs.srid
        return polygon


@dataclass(frozen=True)
class CRS:
    """
    Represents a CRS (Coordinate Reference System), which preferably follows the URN format
    as specified by `the OGC consortium <http://www.opengeospatial.org/ogcUrnPolicy>`_.
    """

    # CRS logic, based upon https://github.com/wglas85/django-wfs/blob/master/wfs/helpers.py
    # Copyright (c) 2006 Wolfgang Glas - Apache 2.0 licensed
    # Ported to Python 3.6 style.

    #: Either "ogc" or "opengis", whereas "ogc" is highly recommended.
    domain: str

    #: Either "OGC" or "EPSG".
    authority: str

    #: The version of the authorities' SRS registry, which is empty or
    #: contains two or three numeric components separated by dots like "6.9" or "6.11.9".
    #: For WFS 2.0 this is typically empty.
    version: str

    #: A string representation of the coordinate system reference ID.
    #: For OGC, only "CRS84" is supported as crsid. For EPSG, this is the formatted CRSID.
    crsid: str

    #: The integer representing the numeric spatial reference ID as
    #: used by the EPSG and GIS database backends.
    srid: int

    @classmethod
    def from_string(cls, uri: Union[str, int]):
        """
        Parse an CRS (Coordinate Reference System) URI, which preferably follows the URN format
        as specified by `the OGC consortium <http://www.opengeospatial.org/ogcUrnPolicy>`_
        and construct a new CRS instance.

        The value can be 3 things:

        * A URI in OGC URN format.
        * A legacy CRS URI ("epsg:<SRID>", or "http://www.opengis.net/...").
        * A numeric SRID (equivalent to "urn:ogc:def:crs:EPSG:6.9:<SRID>")
        """
        if isinstance(uri, int):
            return cls._from_srid(uri)
        elif uri.startswith("urn:"):
            return cls._from_urn(uri)
        else:
            return cls._from_legacy(uri)

    @classmethod
    def _from_srid(cls, srid: int):
        """Instantiate this class using an numeric spatial reference ID"""
        return cls(
            domain="ogc", authority="EPSG", version="", crsid=str(srid), srid=int(srid),
        )

    @classmethod
    def _from_urn(cls, urn):  # noqa: C901
        """Instantiate this class using an URN format."""
        urn_match = CRS_URN_REGEX.match(urn)
        if not urn_match:
            raise ValueError(
                f"Unknown CRS URN [{urn}] specified: {CRS_URN_REGEX.pattern}"
            )

        domain = urn_match.group("domain")
        authority = urn_match.group("authority").upper()

        if domain not in ("ogc", "opengis"):
            raise ValueError(f"CRS URI [{urn}] contains unknown domain [{domain}]")

        if authority == "EPSG":
            crsid = urn_match.group("id")
            try:
                srid = int(crsid)
            except ValueError:
                raise SyntaxError(
                    f"CRS URI [{urn}] should contain a numeric SRID value."
                ) from None
        elif authority == "OGC":
            crsid = urn_match.group("id").upper()
            if crsid != "CRS84":
                raise ValueError(f"OGC CRS URI from [{urn}] contains unknown id [{id}]")
            srid = 4326
        else:
            raise ValueError(
                f"CRS URI [{urn}] contains unknown authority [{authority}]"
            )

        return cls(
            domain=domain,
            authority=authority,
            version=urn_match.group(3),
            crsid=crsid,
            srid=srid,
        )

    @classmethod
    def _from_legacy(cls, uri):
        """Instantiate this class from a legacy URL"""
        luri = uri.lower()
        for head in (
            "epsg:",
            "http://www.opengis.net/def/crs/epsg/0/",
            "http://www.opengis.net/gml/srs/epsg.xml#",
        ):
            if luri.startswith(head):
                crsid = luri[len(head) :]
                try:
                    srid = int(crsid)
                except ValueError:
                    raise SyntaxError(
                        f"CRS URI [{uri}] should contain a numeric SRID value."
                    ) from None

                return cls(
                    domain="ogc", authority="EPSG", version="", crsid=crsid, srid=srid,
                )

        raise ValueError(f"Unknown CRS URI [{uri}] specified")

    @property
    def legacy(self):
        """Return a legacy string in the format "EPSG:<srid>"""
        return f"EPSG:{self.srid:d}"

    @property
    def urn(self):
        """Return The OGC URN corresponding to this CRS. """
        return f"urn:{self.domain}:def:crs:{self.authority}:{self.version or ''}:{self.crsid}"

    def __str__(self):
        return self.urn

    def __eq__(self, other):
        if isinstance(other, CRS):
            # CRS84 is NOT equivalent to EPSG:4326.
            # EPSG:4326 specifies coordinates in lat/long order and CRS:84 in long/lat order.
            return self.authority == other.authority and self.srid == other.srid
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self.authority, self.srid))


WGS84 = CRS.from_string(4326)  # aka EPSG:4326



from gisserver.features import BoundingBox


class TestBoundingBox:
    def test_parse_4(self):
        bbox = BoundingBox.from_string("-1,-2,3,4")
        assert bbox.lower_corner == [-1, -2]
        assert bbox.upper_corner == [3, 4]

    def test_parse_5(self):
        bbox = BoundingBox.from_string("-1,-2,3,4,urn:ogc:def:crs:EPSG::28992")
        assert bbox.lower_corner == [-1, -2]
        assert bbox.upper_corner == [3, 4]
        assert bbox.crs.srid == 28992

    def test_extend(self):
        bbox = BoundingBox(1, 2, 3, 4)
        bbox.extend_to(-4, -3, -2, -1)
        assert bbox.lower_corner == [-4, -3]
        assert bbox.upper_corner == [3, 4]



from doctest import Example
from functools import lru_cache
from pathlib import Path
from typing import Union

from lxml import etree
from lxml.doctestcompare import PARSE_XML, LXMLOutputChecker

# XSD schemas are downloaded from http://schemas.opengis.net/wfs/2.0/wfs.xsd
# using https://github.com/n-a-t-e/xsd_download/blob/master/xsd_download.py
# The download itself is really slow, hence these files are cached.
XSD_ROOT = Path(__file__).parent.parent.joinpath("files/xsd")
WFS_20_XSD = XSD_ROOT.joinpath("schemas.opengis.net/wfs/2.0/wfs.xsd")


@lru_cache(maxsize=100)
def compile_xsd(xsd_file) -> etree.XMLSchema:
    """Compile the XSD files into an lxml """
    xmlschema_doc = etree.parse(str(xsd_file))
    return etree.XMLSchema(xmlschema_doc)


def validate_xsd(xml_text: bytes, xsd_file: str) -> etree._Element:
    """Validate an XML file"""
    xml_schema = compile_xsd(xsd_file)

    xml_doc = etree.fromstring(xml_text)
    if not xml_schema.validate(xml_doc):
        # Improve error message display, to ease debugging of XML data
        source_lines = xml_text.decode().splitlines()
        raise etree.DocumentInvalid(
            "\n".join(
                [
                    f"{err.message} at {err.line}:{err.column} "
                    f"in {err.path} (source: {source_lines[err.line - 1].strip()})"
                    for err in xml_schema.error_log
                ]
            )
        )

    return xml_doc


def assert_xml_equal(got: Union[bytes, str], want: str):
    """Compare two XML strings."""
    checker = LXMLOutputChecker()
    if not checker.check_output(want, got, PARSE_XML):
        example = Example("", "")
        example.want = want  # unencoded, avoid doctest for bytes type.
        message = checker.output_difference(example, got, PARSE_XML)
        raise AssertionError(message)
